# Generated by Django 2.0.3 on 2019-03-28 21:01

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='LastScheduledFor',
            fields=[
                ('task', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('execute_at', models.DateTimeField()),
            ],
        ),
    ]



import time
import logging
import select
import abc

from .models import Job
from django.db import connection, transaction


class Queue(object, metaclass=abc.ABCMeta):
    job_model = Job
    logger = logging.getLogger(__name__)

    def __init__(self, tasks, notify_channel=None):
        self.tasks = tasks
        self.notify_channel = notify_channel

    @abc.abstractmethod
    def run_once(self):
        raise NotImplementedError

    def run_job(self, job):
        task = self.tasks[job.task]
        start_time = time.time()
        retval = task(self, job)
        self.logger.info(
            'Processing %r took %0.4f seconds. Task returned %r.',
            job, time.time() - start_time, retval,
            extra={
                'data': {
                    'job': job.to_json(),
                    'retval': retval,
                }
            }
        )
        return retval

    def enqueue(self, task, args={}, execute_at=None, priority=None):
        assert task in self.tasks
        kwargs = {
            'task': task,
            'args': args,
        }
        if execute_at is not None:
            kwargs['execute_at'] = execute_at
        if priority is not None:
            kwargs['priority'] = priority

        job = self.job_model.objects.create(**kwargs)
        if self.notify_channel:
            self.notify()
        return job

    def listen(self):
        assert self.notify_channel, "You must set a notify channel in order to listen."
        with connection.cursor() as cur:
            cur.execute('LISTEN "{}";'.format(self.notify_channel))

    def wait(self, timeout=30):
        connection.connection.poll()
        notifies = self.filter_notifies()
        if notifies:
            return notifies

        select.select([connection.connection], [], [], timeout)
        connection.connection.poll()
        return self.filter_notifies()

    def filter_notifies(self):
        notifies = [
            i for i in connection.connection.notifies
            if i.channel == self.notify_channel
        ]
        connection.connection.notifies = [
            i for i in connection.connection.notifies
            if i.channel != self.notify_channel
        ]
        return notifies

    def notify(self):
        with connection.cursor() as cur:
            cur.execute('NOTIFY "{}";'.format(self.notify_channel))


class AtMostOnceQueue(Queue):
    def run_once(self, exclude_ids=[]):
        assert not connection.in_atomic_block
        job = None
        try:
            job = self.job_model.dequeue(exclude_ids=exclude_ids)
            if job:
                self.logger.debug('Claimed %r.', job, extra={
                    'data': {
                        'job': job.to_json(),
                    }
                })
                return (job, self.run_job(job), None)
            else:
                return None
        except Exception as e:
            return (job, None, e)


class AtLeastOnceQueue(Queue):
    def run_once(self, exclude_ids=[]):
        job = None
        try:
            with transaction.atomic():
                job = self.job_model.dequeue(exclude_ids=exclude_ids)
                if job:
                    self.logger.debug('Claimed %r.', job, extra={
                        'data': {
                            'job': job.to_json(),
                        }
                    })
                    return (job, self.run_job(job), None)
                else:
                    return None
        except Exception as e:
            return (job, None, e)



import time
import pytz
import logging
import datetime

from django.core.management.base import BaseCommand, CommandError
from django.db import transaction, connection

from dpq_scheduler.models import LastScheduledFor


class Scheduler(BaseCommand):
    # The set of tasks to schedule. dict of task name -> schedule func. A
    # schedule func takes a datetime and returns the soonest datetime in the
    # schedule that is strictly after the argument.
    tasks = None
    # The queue to schedule in.
    queue = None

    logger = logging.getLogger(__name__)

    def handle(self, **options):
        self.validate_tasks()

        while True:
            delay = self.one_round()
            self.logger.debug('Sleeping %s.', delay)
            time.sleep(min(delay.total_seconds(), 2 ** 30))

    @transaction.atomic
    def one_round(self):
        self.logger.debug('Beginning a scheduling round.')
        with connection.cursor() as cursor:
            cursor.execute("LOCK TABLE %s IN EXCLUSIVE MODE" % LastScheduledFor._meta.db_table)
            cursor.execute("SELECT now()")
            now, = cursor.fetchone()

        last_scheduleds = LastScheduledFor.objects.in_bulk()

        next_wakeup = datetime.datetime.max.replace(tzinfo=pytz.UTC)

        for task_name, schedule in self.tasks.items():
            try:
                last_scheduled = last_scheduleds[task_name]
            except KeyError:
                # This isn't actually marking that the task has been scheduled,
                # we just need to create the row for the first time somehow,
                # and this will let the next iteration actually schedule the
                # task.
                LastScheduledFor.objects.create(task=task_name, execute_at=now)
                next_time = schedule(now)
            else:
                next_time = schedule(last_scheduled.execute_at)
                if next_time <= now:
                    task = self.queue.enqueue(task_name, execute_at=next_time)
                    self.logger.info('Scheduled %r for %s.', task, next_time)
                    last_scheduled.execute_at = next_time
                    last_scheduled.save()
                    next_time = schedule(next_time)
            next_wakeup = min(next_wakeup, next_time)

        # these tasks have been removed from the config
        LastScheduledFor.objects.filter(
            task__in=set(last_scheduleds) - set(self.tasks)
        ).delete()

        return max(next_wakeup - now, datetime.timedelta(seconds=0))

    def validate_tasks(self):
        missing_tasks = set(self.tasks.keys()) - set(self.queue.tasks.keys())
        if missing_tasks:
            raise CommandError("Tasks in schedule but missing from queue: %r" % missing_tasks)



from django.db import models


class LastScheduledFor(models.Model):
    task = models.CharField(max_length=255, primary_key=True)
    execute_at = models.DateTimeField()

    def __str__(self):
        return '{}: {}'.format(self.task, self.execute_at)



import datetime
import pytz

from django.test import SimpleTestCase

from .schedules import repeater, every_day_at, every_dow_at


DENVER = pytz.timezone('America/Denver')


def tztime(tz, *args, **kwargs):
    if not isinstance(tz, datetime.tzinfo):
        tz = pytz.timezone(tz)
    naive = datetime.datetime(*args, **kwargs)
    return tz.localize(naive)


class TestSchedulers(SimpleTestCase):
    def test_repeater(self):
        now = tztime('UTC', 2019, 1, 1, 0, 0, 0)
        n = repeater(datetime.timedelta(seconds=5))(now)
        self.assertEqual(n, tztime('UTC', 2019, 1, 1, 0, 0, 5))

    def test_every_day_at_before(self):
        now = tztime('UTC', 2019, 1, 1, 0, 0, 0)
        n = every_day_at(datetime.time(8), pytz.UTC)(now)
        self.assertEqual(n, tztime('UTC', 2019, 1, 1, 8))

    def test_every_day_at_after(self):
        now = tztime('UTC', 2019, 1, 1, 9, 0, 0)
        n = every_day_at(datetime.time(8), pytz.UTC)(now)
        self.assertEqual(n, tztime('UTC', 2019, 1, 2, 8))

    def test_every_day_dst_transition(self):
        now = tztime(DENVER, 2019, 11, 2, 8, 0, 0)
        n = every_day_at(datetime.time(8), DENVER)(now)
        self.assertEqual(n, tztime(DENVER, 2019, 11, 3, 8))

    def test_every_day_different_time_zones(self):
        now = tztime('UTC', 2019, 11, 3, 0, 0, 0)
        n = every_day_at(datetime.time(22), pytz.FixedOffset(-180))(now)
        self.assertEqual(n, tztime(pytz.FixedOffset(-180), 2019, 11, 2, 22))

    def test_every_dow_at_before(self):
        now = tztime(DENVER, 2019, 11, 2, 8, 0, 0)
        n = every_dow_at(6, datetime.time(8), DENVER)(now)
        self.assertEqual(n, tztime(DENVER, 2019, 11, 3, 8))

    def test_every_dow_at_after(self):
        now = tztime(DENVER, 2019, 11, 3, 8, 0, 0)
        n = every_dow_at(2, datetime.time(8), DENVER)(now)
        self.assertEqual(n, tztime(DENVER, 2019, 11, 6, 8))

    def test_every_dow_equal(self):
        now = tztime(DENVER, 2019, 11, 3, 8, 0, 0)
        n = every_dow_at(6, datetime.time(8), DENVER)(now)
        self.assertEqual(n, tztime(DENVER, 2019, 11, 10, 8))

    def test_every_dow_different_time_zones(self):
        now = tztime('UTC', 2019, 11, 4, 0, 0, 0)
        n = every_dow_at(6, datetime.time(22), pytz.FixedOffset(-180))(now)
        self.assertEqual(n, tztime(pytz.FixedOffset(-180), 2019, 11, 3, 22))



# Generated by Django 2.0.3 on 2019-04-19 20:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dpq', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='job',
            name='priority',
            field=models.IntegerField(default=0, help_text='Jobs with higher priority will be processed first.'),
        ),
    ]



# Generated by Django 3.2.5 on 2021-07-16 19:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dpq', '0002_auto_20190419_2057'),
    ]

    if hasattr(models, 'JSONField'):
        operations = [
            migrations.AlterField(
                model_name='job',
                name='args',
                field=models.JSONField(),
            ),
        ]
    else:
        operations = []



import datetime


def combine(date, time, tzinfo):
    """
    Like datetime.datetime.combine, but works across dst transitions.
    """
    return tzinfo.localize(datetime.datetime.combine(date, time))


def repeater(delta):
    return lambda prev: prev + delta


def every_day_at(t, tz):
    def schedule(prev):
        todays_at = combine(prev.astimezone(tz).date(), t, tz)
        if prev < todays_at:
            return todays_at
        else:
            return combine(todays_at + datetime.timedelta(days=1), t, tz)
    return schedule


def every_dow_at(dow, at, tz):
    def schedule(prev):
        prev = prev.astimezone(tz)
        start_of_this_week = prev.date() - datetime.timedelta(days=prev.date().weekday())
        this_weeks_at = combine(start_of_this_week + datetime.timedelta(days=dow), at, tz)
        if prev < this_weeks_at:
            return this_weeks_at
        else:
            return combine(start_of_this_week + datetime.timedelta(days=7 + dow), at, tz)
    return schedule



import logging
import signal
import time
import os

from django.core.management.base import BaseCommand
from django.db import connection


class Worker(BaseCommand):
    # The queue to process. Subclass and set this.
    queue = None
    logger = logging.getLogger(__name__)

    def add_arguments(self, parser):
        parser.add_argument(
            '--delay',
            type=float,
            default=1,
            help="The number of seconds to wait to check for new tasks.",
        )
        parser.add_argument(
            '--listen',
            action='store_true',
            help="Use LISTEN/NOTIFY to wait for events."
        )

    def handle_shutdown(self, sig, frame):
        if self._in_task:
            self.logger.info('Waiting for active tasks to finish...')
            self._shutdown = True
        else:
            raise InterruptedError

    def run_available_tasks(self):
        """
        Runs tasks continuously until there are no more available.
        """
        # Prevents tasks that failed from blocking others.
        failed_tasks = set()
        while True:
            self._in_task = True
            result = self.queue.run_once(exclude_ids=failed_tasks)
            job, retval, exc = result or (None, None, None)
            if exc:
                if job:
                    self.logger.exception('Error in %r: %r.', job, exc, extra={
                        'data': {
                            'job': job.to_json(),
                        },
                    })
                    failed_tasks.add(job.id)
                else:
                    # This is an exception before a task could even be
                    # retrieved, so it's probably fatal
                    raise exc
            self._in_task = False
            if self._shutdown:
                raise InterruptedError
            if not job:
                break

    def handle(self, **options):
        self._shutdown = False
        self._in_task = False

        self.delay = options['delay']
        self.listen = options['listen']

        with connection.cursor() as cursor:
            cursor.execute("SET application_name TO %s", ['dpq#{}'.format(os.getpid())])

        if self.listen:
            self.queue.listen()
        try:
            # Handle the signals for warm shutdown.
            signal.signal(signal.SIGINT, self.handle_shutdown)
            signal.signal(signal.SIGTERM, self.handle_shutdown)

            while True:
                self.run_available_tasks()
                self.wait()
        except InterruptedError:
            # got shutdown signal
            pass

    def wait(self):
        if self.listen:
            count = len(self.queue.wait(self.delay))
            self.logger.debug('Woke up with %s NOTIFYs.', count)
            return count
        else:
            time.sleep(self.delay)
            return 1



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-02 18:45
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
import django.contrib.postgres.functions
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Job',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(default=django.contrib.postgres.functions.TransactionNow)),
                ('execute_at', models.DateTimeField(default=django.contrib.postgres.functions.TransactionNow)),
                ('priority', models.PositiveIntegerField(default=0, help_text='Jobs with higher priority will be processed first.')),
                ('task', models.CharField(max_length=255)),
                ('args', django.contrib.postgres.fields.jsonb.JSONField()),
            ],
        ),
        migrations.AddIndex(
            model_name='job',
            index=models.Index(fields=['-priority', 'created_at'], name='dpq_job_priorit_aa4927_idx'),
        ),
    ]



import datetime
import logging
import random

from django.db import transaction


def repeat(delay):
    """
    Endlessly repeats a task, every `delay` (a timedelta).

    Under at-least-once delivery, the tasks can not overlap. The next scheduled
    task only becomes visible once the previous one commits.

        @repeat(datetime.timedelta(minutes=5)
        def task(queue, job):
            pass

    This will run `task` every 5 minutes. It's up to you to kick off the first
    task, though.
    """
    def decorator(fn):
        def inner(queue, job):
            queue.enqueue(
                job.task,
                job.args,
                execute_at=job.execute_at + delay,
                priority=job.priority,
            )
            return fn(queue, job)

        return inner
    return decorator


def exponential_with_jitter(offset=6):
    def delayfn(retries):
        jitter = random.randrange(-15, 15)
        return datetime.timedelta(seconds=2 ** (retries + offset) + jitter)
    return delayfn


def retry(max_retries, delayfn=exponential_with_jitter(), Exc=Exception):
    def decorator(fn):
        logger = logging.getLogger(__name__)

        def inner(queue, job):
            try:
                with transaction.atomic():
                    return fn(queue, job)
            except Exc as e:
                retries = job.args.get('retries', 0)
                if retries < max_retries:
                    job.args['retries'] = retries + 1
                    delay = delayfn(retries)
                    job.execute_at += delay
                    job.save(force_insert=True)
                    logger.warning(
                        'Task %r failed: %s. Retrying in %s.', job, e, delay, exc_info=True,
                    )
                else:
                    logger.exception(
                        'Task %r exceeded its retry limit: %s.', job, e, exc_info=True,
                    )
        return inner
    return decorator



from django.db import models
from django.db import connection
from django.contrib.postgres.functions import TransactionNow
try:
    from django.db.models import JSONField
except ImportError:
    # django < 3.1
    from django.contrib.postgres.fields import JSONField


class BaseJob(models.Model):
    id = models.BigAutoField(primary_key=True)
    created_at = models.DateTimeField(default=TransactionNow)
    execute_at = models.DateTimeField(default=TransactionNow)
    priority = models.IntegerField(
        default=0,
        help_text="Jobs with higher priority will be processed first."
    )
    task = models.CharField(max_length=255)
    args = JSONField()

    class Meta:
        indexes = [
            models.Index(fields=['-priority', 'created_at']),
        ]
        abstract = True

    def __str__(self):
        return '%s: %s' % (self.id, self.task)

    @classmethod
    def dequeue(cls, exclude_ids=[]):
        """
        Claims the first available task and returns it. If there are no
        tasks available, returns None.

        For at-most-once delivery, commit the transaction before
        processing the task. For at-least-once delivery, dequeue and
        finish processing the task in the same transaction.

        To put a job back in the queue, you can just call
        .save(force_insert=True) on the returned object.
        """

        tasks = list(cls.objects.raw(
            """
            DELETE FROM {db_table}
            WHERE id = (
                SELECT id
                FROM {db_table}
                WHERE execute_at <= now()
                  AND NOT id = ANY(%s)
                ORDER BY priority DESC, created_at
                FOR UPDATE SKIP LOCKED
                LIMIT 1
            )
            RETURNING *;
            """.format(
                db_table=connection.ops.quote_name(cls._meta.db_table),
            ),
            [list(exclude_ids)]
        ))
        assert len(tasks) <= 1
        if tasks:
            return tasks[0]
        else:
            return None

    def to_json(self):
        return {
            'id': self.id,
            'created_at': self.created_at,
            'execute_at': self.execute_at,
            'priority': self.priority,
            'task': self.task,
            'args': self.args,
        }

class Job(BaseJob):
    pass



from setuptools import setup


setup(
    name='django-postgres-queue',
    version='1.0.0',
    packages=[
        'dpq',
        'dpq.migrations',
        'dpq_scheduler',
        'dpq_scheduler.migrations',
    ],
    license='BSD',
    long_description=open('README.rst').read(),
    author="Gavin Wahl",
    author_email="gavinwahl@gmail.com",
    url="https://github.com/gavinwahl/django-postgres-queue",
    install_requires=[
        'Django>=1.11',
    ]
)
